;; Generated package description from elisa.el  -*- no-byte-compile: t -*-
(define-package "elisa" "1.1.7.0.20250202.204048" "Emacs Lisp Information System Assistant" '((emacs "29.2") (ellama "0.11.2") (llm "0.18.1") (async "1.9.8") (plz "0.9")) :commit "b655b59d371639d357dcabe48f1c2cd1694ee8de" :authors '(("Sergey Kostyaev" . "sskostyaev@gmail.com")) :maintainer '("Sergey Kostyaev" . "sskostyaev@gmail.com") :keywords '("help" "local" "tools") :url "http://github.com/s-kostyaev/elisa")
